#include "chrono.h"
#include <stdio.h>
#include <conio.h>
#include <stdarg.h>
#include <ctype.h>
#include <memory.h>

Console::Console()
{ Screen = NULL;
  hRead  = GetStdHandle(STD_INPUT_HANDLE);
  hWrite = GetStdHandle(STD_OUTPUT_HANDLE);
  Width  = Height = 0;
}

Console::~Console()
{ delete[] Screen;
}

void Console::showCursor(bool show, DWORD size)
{ CONSOLE_CURSOR_INFO ci;
  ci.dwSize   = size;
  ci.bVisible = show;
  SetConsoleCursorInfo(hWrite, &ci);
}

void Console::gotoXY(int x, int y)
{ COORD pos = {x, y};
  SetConsoleCursorPosition(hWrite, pos);
}

void Console::fromXY(int *x, int *y)
{ CONSOLE_SCREEN_BUFFER_INFO si;
  GetConsoleScreenBufferInfo(hWrite, &si);
  *x = si.dwCursorPosition.X, *y = si.dwCursorPosition.Y;
}
  
void Console::setTitle(const char *str)
{ SetConsoleTitle(str);
}

bool Console::setSize(int wid, int hei)
{ COORD size = {wid, hei};
  bool ret;
  if(wid<Width || hei<Height)
  { SMALL_RECT rect = {0, 0};
    rect.Right = wid-1;
    rect.Bottom = hei-1;
    ret = (SetConsoleWindowInfo(hWrite, TRUE, &rect) == TRUE);
  }
  ret = (SetConsoleScreenBufferSize(hWrite, size) == TRUE);
  if(ret)
  { delete[] Screen;
    Screen = new CHAR_INFO[wid*hei];
    hRead  = GetStdHandle(STD_INPUT_HANDLE);
    hWrite = GetStdHandle(STD_OUTPUT_HANDLE);
    Mode   = ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT;
    SetConsoleMode(hWrite, ENABLE_WRAP_AT_EOL_OUTPUT | ENABLE_PROCESSED_OUTPUT);
  }
  if(ret && wid>=Width && hei>=Height)
  { SMALL_RECT rect = {0, 0};
    rect.Right = wid-1;
    rect.Bottom = hei-1;
    ret = (SetConsoleWindowInfo(hWrite, TRUE, &rect) == TRUE);
  }
  if(ret) Width = wid, Height = hei;
  return ret;
}

int Console::getS(char *buf, int maxlen)
{ DWORD dummy;
  if(Mode != (ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_PROCESSED_INPUT))
    SetConsoleMode(hRead, Mode = ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT |
                   ENABLE_PROCESSED_INPUT);
  ReadConsole(hRead, buf, maxlen, &dummy, NULL);
  getch();
  buf[dummy] = 0;
  trim(buf);
  return strlen(buf);
}

int Console::putS(const char *str)
{ DWORD ret;
  WriteConsole(hWrite, str, strlen(str), &ret, NULL);
  return (int)ret;
}

int Console::printF(const char *str, ...)
{ char buf[MAXLINE];
  va_list list;
  va_start(list, str);
  int len = vsprintf(buf, str, list);
  va_end(list);
  return putS(buf);
}

char Console::getch()
{ DWORD dummy;
  char c;
  if(Mode != 0) SetConsoleMode(hRead, Mode=0);
  ReadConsole(hRead, &c, 1, &dummy, NULL);
  return c;
}

char Console::getche()
{ DWORD dummy;
  char c;
  if(Mode != ENABLE_ECHO_INPUT) SetConsoleMode(hRead, Mode=ENABLE_ECHO_INPUT);
  ReadConsole(hRead, &c, 1, &dummy, NULL);
  return c;
}

char Console::putch(char c)
{ DWORD dummy;
  WriteConsole(hWrite, &c, 1, &dummy, NULL);
  return c;
}

byte Console::charAt(int x, int y)
{ COORD pos = {x,y};
  DWORD dummy;
  char  ret;
  ReadConsoleOutputCharacter(hWrite, &ret, 1, pos, &dummy);
  return ret;
}

byte Console::attrAt(int x, int y)
{ COORD pos = {x,y};
  DWORD dummy;
  WORD  ret;
  ReadConsoleOutputAttribute(hWrite, &ret, 1, pos, &dummy);
  return (byte)ret;
}

CharData Console::bothAt(int x, int y)
{ COORD pos = {x,y};
  DWORD    dummy;
  WORD     attr;
  CharData ret;
  ReadConsoleOutputCharacter(hWrite, (char*)&ret.b.c, 1, pos, &dummy);
  ReadConsoleOutputAttribute(hWrite, &attr, 1, pos, &dummy);
  ret.b.attr = (byte)attr;
  return ret;
}

byte Console::charTo(int x, int y, byte c)
{ COORD pos = {x, y};
  DWORD dummy;
  WriteConsoleOutputCharacter(hWrite, (char*)&c, 1, pos, &dummy);
  return c;
}

byte Console::attrTo(int x, int y, byte attr)
{ COORD pos = {x, y};
  DWORD dummy;
  WORD  bleh = attr;
  WriteConsoleOutputAttribute(hWrite, &bleh, 1, pos, &dummy);
  return attr;
}

byte Console::bothTo(int x, int y, byte c, byte attr)
{ COORD pos = {x, y};
  DWORD dummy;
  WORD  bleh = attr;
  WriteConsoleOutputCharacter(hWrite, (char*)&c,    1, pos, &dummy);
  WriteConsoleOutputAttribute(hWrite, &bleh, 1, pos, &dummy);
  return c;
}

byte Console::setColor(byte attr)
{ SetConsoleTextAttribute(hWrite, attr);
  return attr;
}

byte Console::getColor()
{ CONSOLE_SCREEN_BUFFER_INFO si;
  GetConsoleScreenBufferInfo(hWrite, &si);
  return (byte)si.wAttributes;
}

void Console::fill(int x, int y, int wid, int hei, byte c, byte attr)
{ DWORD dummy;
  COORD st   = {x,y};
  int   e    = y+hei;
  WORD  bleh = attr;
  for(;st.Y<e;st.Y++)
  { FillConsoleOutputCharacter(hWrite, c, wid, st, &dummy);
    FillConsoleOutputAttribute(hWrite, bleh, wid, st, &dummy);
  }
}

void Console::putBlock(int x, int y, int wid, int hei, const byte *buf)
{ COORD pos = {0, 0}, size = {wid, hei};
  SMALL_RECT drect = {x, y, x+wid-1, y+hei-1};
  int i, len=wid*hei;
  for(i=0;i<len;i++)
  { Screen[i].Char.AsciiChar = (char)buf[i*2];
    Screen[i].Attributes     = (WORD)buf[i*2+1];
  }
  SetConsoleMode(hWrite, 0);
  WriteConsoleOutput(hWrite, Screen, size, pos, &drect);
  SetConsoleMode(hWrite, ENABLE_WRAP_AT_EOL_OUTPUT | ENABLE_PROCESSED_OUTPUT);
}

void Console::getBlock(int x, int y, int wid, int hei, byte *buf)
{ COORD pos = {0, 0}, size = {wid, hei};
  SMALL_RECT srect = {x, y, x+wid-1, y+hei-1};
  ReadConsoleOutput(hWrite, Screen, size, pos, &srect);
  int i, len=wid*hei;
  for(i=0;i<len;i++)
  { buf[i*2]   = (byte)Screen[i].Char.AsciiChar;
    buf[i*2+1] = (byte)Screen[i].Attributes;
  }
}

BOOL WINAPI Console::CtrlHandler(DWORD ctrl)
{ if(ctrl == CTRL_C_EVENT || ctrl == CTRL_BREAK_EVENT) return TRUE;
  return FALSE;
}
