#ifndef _CONSOLE_H
#define _CONSOLE_H

#include <windows.h>
#include "datatype.h"
#include "config.h"

#define CHARIND(x,y)      (((y)*MAPWID+(x))*2)
#define CHARAT(x,y,buf)   ((buf)[CHARIND(x,y)])
#define CHARTO(x,y,buf,c)  (buf)[CHARIND(x,y)] = (c);
#define ATTRAT(x,y,buf)   ((buf)[CHARIND(x,y)+1])
#define ATTRTO(x,y,buf,c)  (buf)[CHARIND(x,y)+1] = (c);
#define TOBACKCLR(c)      ((c) << 4)

class Console
{ public:
    Console();
    ~Console();

    void     showCursor(bool show, DWORD size=20); // size is 1 to 99
    void     gotoXY(int x, int y);
    void     fromXY(int *x, int *y);
    void     setTitle(const char *str);
    bool     setSize(int wid, int hei);       // returns success flag

    int      getS(char *buf, int maxlen);     // returns string length
    int      putS(const char *str);           // returns # chars printed
    int      printF(const char *str, ...);    // returns # chars printed
    
    char     getch();
    char     getche();
    char     putch(char c);                   // returns char passed

    byte     charAt(int x, int y);
    byte     attrAt(int x, int y);
    CharData bothAt(int x, int y);

    byte     charTo(int x, int y, byte c);    // returns char passed
    byte     attrTo(int x, int y, byte attr); // returns attr passed
    byte     bothTo(int x, int y, byte c, byte attr); // returns char passed
    byte     bothTo(int x, int y, CharData cd)// returns char passed
    { return bothTo(x, y, cd.b.c, cd.b.attr); }

    byte     setColor(byte attr);             // returns attr passed
    byte     getColor();

    void     fill(int x, int y, int wid, int hei, byte c, byte attr);
    void     fill(int x, int y, int wid, int hei, CharData cd)
    { fill(x, y, wid, hei, cd.b.c, cd.b.attr); }

    void     putBlock(int x, int y, int wid, int hei, const byte *buf);
    void     getBlock(int x, int y, int wid, int hei, byte *buf);

    int      Width, Height;
  private:
    CHAR_INFO *Screen;
    HANDLE     hRead, hWrite;
    DWORD      Mode;

    static BOOL WINAPI CtrlHandler(DWORD);
};

#endif
