#include "chrono.h"
#include <stdarg.h>
#include <stdio.h>
#include <new.h>

#define BEGIN_CREATLIST() { int index=0;
#define END_CREATLIST()   for(int i=0;i<IT_NUMTYPES;i++) \
                          creatFact[i].flags = 0, creatFact[i].idTag = -1; }
#define BEGIN_CLASS(cls)  CreatClasses[cls].start = index;
#define END_CLASS(cls)    CreatClasses[cls].end   = index-1;
#define ADD_CREAT(cre)    new (creatFact+(cre::TYPE)) cre; index++;


CharData CreatChars[CT_NUMTYPES] = 
{ {'@', WHITE}
};

CreatClass CreatClasses[CC_NUMCLASS] =
{ {"player"}
};

static Creature  creatFact[CT_NUMTYPES];
static Creature *creatStore;
static int       numCreats;
static int       creatCache[CREATCACHE];

static int findFreeCreature()
{ int i;
  for(i=CREATCACHE-1; i>=0; i--)
    if(!(creatStore[creatCache[i]].flags & CF_INSTORE)) return creatCache[i];
  for(i=0; i<numCreats; i++) if(!(creatStore[i].flags & CF_INSTORE)) return i;

  i = numCreats;
  Creature *store = new Creature[numCreats=numCreats*2];
  memcpy(store, creatStore, i*sizeof(Creature));
  creatStore = store;
  for(int j=i;j<numCreats;j++) creatStore[j].flags &= ~CF_INSTORE;
  return i;
}

static Creature * getFreeCreature()
{ int       pos  = findFreeCreature();
  Creature *ret  = creatStore+pos;
  ret->idTag     = pos;
  return ret;
}

static void releaseCreature(Creature *creat)
{ creat->flags = 0;
  memmove(creatCache, creatCache+1, (CREATCACHE-1)*sizeof(int));
  creatCache[CREATCACHE-1] = creat->idTag;
}

int initCreatures()
{ // the creature store
  creatStore = new Creature[numCreats=32];
  for(int i=0;i<numCreats;i++) creatStore[i].flags &= ~CF_INSTORE;

  // individual items (classes must be in order defined in the enum!)
  BEGIN_CREATLIST();

  /** CC_PLAYER CLASS **/
  BEGIN_CLASS(CC_PLAYER);
    ADD_CREAT(Player);
  END_CLASS(CC_PLAYER);
  /** END CC_PLAYER CLASS **/

  END_CREATLIST();
  return 0;
}

Creature * newCreat(int type, int x, int y)
{ Creature *creat = &creatFact[type];
  creat = creat->clone();
  creat->pos = Coord(x, y);
  return creat;
}

Creature * newCreatC(int cls, int x, int y)
{ return newCreat(Rnd.Rand(CreatClasses[cls].start, CreatClasses[cls].end), x, y);
}

/***** CREATURE *****/
Creature::Creature()
{ flags  = CF_INSTORE;
  type   = clarse = -1;
  name   = NULL;
  buffer = NULL;
  buflen = 0;
}

Creature::~Creature()
{ if(buffer) freeTempStrings(TS_PLAYER);
  if(name)   delete[] name;
  if(flags & CF_INSTORE) releaseCreature(this);
}

int Creature::load(FILE *file)
{ return 0;
}

int Creature::save(FILE *file)
{ return 0;
}

bool Creature::move(int dir)
{ int nx = pos.x+DirMap[dir].x, ny = pos.y+DirMap[dir].y;
  if(Lev.Map.passable(nx, ny))
  { last = pos;
    pos  = Coord(nx, ny);
    if(isPlayer()) Lev.Map.updateVis(this);
    return true;
  }
  return false;
}

bool Creature::moveFarS(int dir)
{ bool ret=false;
  while(move(dir)) ret=true;
  return ret;
}

bool Creature::open(int dir)
{ int  nx, ny;
  byte b,  type;
  if(dir==DIRSELF) goto nothing;
  nx = pos.x+DirMap[dir].x, ny = pos.y+DirMap[dir].y;

  b  = Lev.Map.layAt(nx, ny);
  if(!(type=Lev.Map.isDoor(b)))
  { nothing:
    tell1("You see nothing you can open there.");
  }
  else
  { if(type == D_OPEN) tell1("It's already open!");
    else
    { Lev.Map.layTo(nx, ny, b==MT_HCLOSED ? MT_HOPEN : MT_VOPEN);
      if(isPlayer()) Lev.Map.updateVis(this);
      tell1("The door opens.");
      return true;
    }
  }
  return false;
}

bool Creature::close(int dir)
{ int  nx, ny;
  byte b,  type;
  if(dir==DIRSELF) goto nothing;
  nx = pos.x+DirMap[dir].x, ny = pos.y+DirMap[dir].y;

  b  = Lev.Map.layAt(nx, ny);
  if(!(type=Lev.Map.isDoor(b)))
  { nothing:
    tell1("You see nothing to close there.");
  }
  else
  { if(type == D_CLOSED) tell1("It's already closed!");
    else
    { Lev.Map.layTo(nx, ny, b==MT_HOPEN ? MT_HCLOSED : MT_VCLOSED);
      if(isPlayer()) Lev.Map.updateVis(this);
      return true;
    }
  }
  return false;
}

bool Creature::pickup(Item *item, Item **src)
{ char let = freeLetter(item);
  if(!let)
  { tell1("Your knapsack cannot accomodate any more items!");
    return false;
  }
  ::remItem(src, item);
  addItem(item);
  tell("%c - %s", item->letter=let, item->ysname());
  return true;
}

bool Creature::drop(Item *item, int count)
{ Item *it = Lev.addItem(pos.x, pos.y, remItem(item, count));
  tell("Dropped %c - %s.", it->letter, it->ysname());
  return true;
}

char Creature::freeLetter(Item *item)
{ Item *it = inv;
  byte lets[52], *found;
  memset(lets, 0, 52);

  while(it)
  { if(item && it->stackable(item)) return it->letter;
    lets[letToInd(it->letter)] = 1;
    it = it->next;
  }
  found = (byte *)memchr(lets, 0, 52);
  if(found) return indToLet(found-lets);
  else return 0;
};

void Creature::addToBuffer(const char *str, int len)
{ assert(isPlayer());
  if(len == -1) len = strlen(str);
  if(!buffer) buffer = getTempString(TS_PLAYER, MAXLINE);
  if(buflen+len >= MAXLINE) flush();
  if(buflen+len >= MAXLINE) print1(str);
  else
  { if(buflen)
    { strcat(buffer, "; ");
      buflen += 2;
    }
    strcat(buffer, str);
    buflen += len;
  }
}

void Creature::flush()
{ if(buflen)
  { print1(buffer);
    buffer[0] = 0;
    buflen = 0;
  }
}

void Creature::tell1(const char *str)
{ if(!isPlayer()) return;
  addToBuffer(str);
}

void Creature::tell(const char *str, ...)
{ if(!isPlayer()) return;
  char buf[MAXLINE];
  va_list list;
  va_start(list, str);
  addToBuffer(buf, vsprintf(buf, str, list));
  va_end(list);
}
/***** END CREATURE *****/

/***** PLAYER *****/
Player::Player()
{ type      = CT_PLAYER;
  clarse    = CC_PLAYER;
  lightrad  = 1;
  sightdist = 5;
}

Creature * Player::clone()
{ return new (getFreeCreature()) Player;
}
/***** END PLAYER *****/
