#include "chrono.h"
#include <ctype.h>
#include <assert.h>

bool doInput()
{ static byte movet = 0, last=0;
  byte c = Con.getch(), act = 0;

  if(bTxtOnTop) clearOut(); 

  switch(c)
  { case K_UP: case K_DOWN: case K_RIGHT: case K_LEFT:
    case K_UPLEFT: case K_UPRIGHT: case K_DNLEFT: case K_DNRIGHT:
      switch(movet)
      { case 0:          act = doMove(c);     break;
        case K_MOVEFARS: act = doMoveFarS(c); break;
      }
      movet = 0;
      break;

    case K_MOVEFARS:
      movet = c; break;

    case K_PICKUP: act = doPickup(); break;
    case K_DROP:   act = doDrop();   break;

    case K_OPEN:   act = doOpen();   break;
    case K_CLOSE:  act = doClose();  break;

    case K_GODOWN: act = doGoDown(); break;
    case K_GOUP:   act = doGoUp();   break;

    case K_RECALL: doRecall(last!=c); break;

    case K_QUIT:   act = doQuit();   break;
  }
  last = c;
  Plyr.flush();

  if(act == QUITGAME) return false;
  if(act == RERENDER) Lev.render();
  if(act) Lev.draw();
  return true;
}

int keyToDir(byte c)
{ switch(c)
  { case K_UP:      return UP;
    case K_DOWN:    return DOWN;
    case K_LEFT:    return LEFT;
    case K_RIGHT:   return RIGHT;
    case K_UPLEFT:  return UPLEFT;
    case K_DNLEFT:  return DNLEFT;
    case K_UPRIGHT: return UPRIGHT;
    case K_DNRIGHT: return DNRIGHT;
    case K_DIRSELF: return DIRSELF;
    default:        return DIRINV;
  }
}

int letToInd(char let)
{ if(islower(let)) return let-'a';
  return let-'A'+26;
}

char indToLet(int ind)
{ if(ind<26) return 'a'+ind;
  return 'A'+ind-26;
}

byte doPrompt(char const *prompt, char const *sel, byte def, bool cs)
{ char buf[MAXLINE], *found;
  byte c;

  clearOut();
  saveCursor();
  strcpy(buf, prompt);
  if(sel)
  { strcat(buf, " [");
    strcat(buf, sel);
    strcat(buf, "]");
  }
  printLine(buf, false);

  c = Con.getch();
  clearOut();
  restCursor();

  if(!sel) return c;
  if(!cs) 
  { if(!(found = strchr(sel, tolower(c)))) found = strchr(sel, toupper(c));
  }
  else found = strchr(sel, c);
  if(!found) return def;
  if(!cs) return toupper(*found);
  else return *found;
}

int askDir()
{ int dir = keyToDir(doPrompt("Which direction?"));
  if(dir == DIRINV) Plyr.tell1("What a silly direction!");
  return dir;
}

void askItems(Item *root, DWORD mflags, ItemClass *classes, int numcs)
{ RECT wnd = {50, 3, 78, 45};
  ItemEnumS ies;
  ies.root   = root;
  ies.cls    = classes ? classes : ItemClasses;
  ies.numcls = classes ? numcs   : IC_NUMCLASS;
  doMenu(&wnd, mflags, (DWORD)&ies, ItemEnum);
}
