#include "chrono.h"
#include <new.h>

#define BEGIN_ITEMLIST() { int index=0;
#define END_ITEMLIST()   for(int i=0;i<IT_NUMTYPES;i++) \
                          itemFact[i].flags = 0, itemFact[i].idTag = -1; }
#define BEGIN_CLASS(cls) ItemClasses[cls].start = index;
#define END_CLASS(cls)   ItemClasses[cls].end   = index-1;
#define ADD_ITEM(itm)    new (itemFact+(itm::TYPE)) itm; index++;

CharData ItemChars[IT_NUMTYPES] = 
{ {'%', RED}
};

ItemClass ItemClasses[IC_NUMCLASS] =
{ {"dummy object"}
};


static Item  itemFact[IT_NUMTYPES];
static Item *itemStore;
static int   numItems;
static int   itemCache[ITEMCACHE];
static char  OutBuf[MAXLINE];

enum
{ YS_A, YS_AN, YS_PAIR, YS_PLURAL
};
static const char * youSees[] =
{ "a ", "an ", "a pair of ",
  // plurals
  "", "", "pairs of "
};
enum
{ PS_NONE, PS_S, PS_ES
};
static const char * plurSufs[] =
{ "", "s", "es"
};

static int findFreeItem()
{ int i;
  for(i=ITEMCACHE-1; i>=0; i--)
    if(!(itemStore[itemCache[i]].flags & IF_INSTORE)) return itemCache[i];
  for(i=0; i<numItems; i++) if(!(itemStore[i].flags & IF_INSTORE)) return i;

  i = numItems;
  Item *store = new Item[numItems=numItems*3/2];
  memcpy(store, itemStore, i*sizeof(Item));
  itemStore = store;
  for(int j=i;j<numItems;j++) itemStore[j].flags &= ~IF_INSTORE;
  return i;
}
  
Item * getFreeItem()
{ int   pos  = findFreeItem();
  Item *ret  = itemStore+pos;
  ret->idTag = pos;
  return ret;
}
  
static void releaseItem(Item *item)
{ item->flags = 0;
  memmove(itemCache, itemCache+1, (ITEMCACHE-1)*sizeof(int));
  itemCache[ITEMCACHE-1] = item->idTag;
}

int initItems()
{ // the item store
  itemStore = new Item[numItems=512];
  for(int i=0;i<numItems;i++) itemStore[i].flags &= ~IF_INSTORE;
  // individual items (classes must be in order defined in the enum!)
  BEGIN_ITEMLIST();

  /** IC_DUMMY CLASS **/
  BEGIN_CLASS(IC_DUMMY);
    ADD_ITEM(IDummy);
  END_CLASS(IC_DUMMY);
  /** END IC_DUMMY CLASS **/

  END_ITEMLIST();
  return 0;
}

Item *addItem(Item **dest, Item *src)
{ assert(!src->prev && !src->next);
  Item *it = *dest;
  while(it)
  { if(src->stackable(it))
    { it->count += src->count;
      releaseItem(src);
      return NULL;
    }
    it = it->next;
  }
  it = *dest;
  if(it)
  { src->next = it;
    it->prev = src;
  }
  return *dest = src;
}

Item *remItem(Item **dest, Item *src, int count)
{ if(count && count < src->count)
  { Item *it = getFreeItem();
    *it = *src;
    it->next    = it->prev = NULL;
    it->count   = count;
    src->count -= count;
    return it;
  }
  if(src->next) src->next->prev = src->prev;
  if(*dest == src) *dest = src->next;
  else src->prev->next = src->next;
  src->next = src->prev = NULL;
  return src;
}

Item * newItem(int type, int count)
{ Item *it = &itemFact[type]; // force call through pointer
  it = it->clone();
  it->count = count;
  return it;
}

Item * newItemC(int cls, int count)
{ return newItem(Rnd.Rand(ItemClasses[cls].start, ItemClasses[cls].end), count);
}

bool ItemEnum(MenuItem *mi, DWORD pies)
{ static int ci;
  static Item *it;
  ItemEnumS *ies = (ItemEnumS *)pies;
  char const *str;

  if(mi->flags & MF_FIRST)
  { ci = 0, it = ies->root, mi->str = const_cast<char *>(ies->cls[ci].name);
    mi->flags |= MF_HEADER;
    return true;
  }
  else if(mi->flags & MF_LAST)
  { freeTempStrings(TS_ITEMENUM);
    return false;
  }

  while(true)
  { while(it && it->clarse != ci) it = it->next;
    if(it) break;
    it = ies->root;
    ci++;
    if(ci < ies->numcls)
    { mi->str = const_cast<char *>(ies->cls[ci].name);
      mi->flags |= MF_HEADER;
      return true;
    }
    else return false;
  }

  str = it->ysname();
  strcpy(mi->str = getTempString(TS_ITEMENUM, strlen(str)), str);
  mi->assoc  = it;
  mi->letter = it->letter;
  it         = it->next;
  return true;
}

/**** ITEM ****/
Item::Item()
{ prev   = next = NULL;
  name   = NULL;
  flags  = IF_INSTORE;
  type   = clarse = -1;
  letter = 0;
}

Item::~Item()
{ if(flags & IF_INSTORE) releaseItem(this);
}

int Item::load(FILE *file)
{ return 0;
}

int Item::save(FILE *file)
{ return 0;
}

const char * Item::ysname()
{ bool desc=false;
  if(count>1)
  { wsprintf(OutBuf, "%d ", count);
    strcat(OutBuf, ysplur);
    strcat(OutBuf, defname);
    strcat(OutBuf, plursuf);
  }
  else
  { strcpy(OutBuf, yousee);
    strcat(OutBuf, defname);
  }
  if(itemFact[type].name)
  { strcat(OutBuf, " described as ");
    strcat(OutBuf, itemFact[type].name);
    desc = true;
  }
  if(name)
  { if(desc) strcat(OutBuf, " and");
    strcat(OutBuf, " named ");
    strcat(OutBuf, name);
  }
  return OutBuf;
}
/**** END ITEM ****/

/**** DUMMY ****/
IDummy::IDummy()
{ type    = IT_DUMMY;
  clarse  = IC_DUMMY;
  yousee  = youSees[YS_A];
  ysplur  = youSees[YS_A+YS_PLURAL];
  plursuf = plurSufs[PS_S];
  defname = "dummy object";
}

Item * IDummy::clone(void *ptr)
{ if(ptr) return new (ptr) IDummy;
  else return new (getFreeItem()) IDummy;
}
/**** END DUMMY ****/
