#include "chrono.h"
#include "level.h"

Level::Level()
{ dlvl = 0;
}

Level::~Level()
{
}

int Level::init()
{ if(Map.init()) return 1;
  memset(Items, 0, sizeof(Items));
  numCreats = 0;
  return 0;
}

void Level::deinit()
{ int i;
  Map.deinit();
  for(i=0;i<numCreats;i++) Creats[i].~Creature();
  numCreats = 0;
  for(i=0;i<MAPWID*MAPHEI;i++)
  { if(Items[i])
    { Item *it = Items[i];
      while(it)
      { Item *next = it->next;
        it->~Item();
        it = next;
      }
      Items[i] = NULL;
    }
  }
}

int Level::load(FILE *file)
{ int i, numItems;
  fread(&dlvl, sizeof(dlvl), 1, file);
  Map.load(file);
  fread(&numCreats, sizeof(numCreats), 1, file);
  for(i=0;i<numCreats;i++) Creats[i].load(file);
  fread(&numItems, sizeof(numItems), 1, file);
  for(i=0;i<numItems;i++)
  { Item *it = getFreeItem();
    it->load(file);
    addItem(it->pos.x, it->pos.y, it);
  }
  fread(&Plyr.pos, sizeof(Plyr.pos), 1, file);
  return 0;
}

int Level::save(FILE *file)
{ int i, numitems=0, pos;
  fwrite(&dlvl, sizeof(dlvl), 1, file);
  Map.save(file);
  fwrite(&numCreats, sizeof(numCreats), 1, file);
  for(i=0;i<numCreats;i++) Creats[i].save(file);
  pos = ftell(file);
  fwrite(&pos, sizeof(int), 1, file);
  for(i=0;i<MAPWID*MAPHEI;i++)
  { if(Items[i])
    { Item *it = Items[i];
      while(it)
      { numitems++;
        it->save(file);
        it = it->next;
      }
    }
  }
  fwrite(&Plyr.pos, sizeof(Plyr.pos), 1, file);
  fseek(file, pos, SEEK_SET);
  fwrite(&numitems, sizeof(int), 1, file);
  fseek(file, 0, SEEK_END);
  return 0;
}

int Level::random()
{ if(GenDungeon()) return 1;
  for(int i=0;i<Map.numStairs;i++)
  { if(Map.Stairs[i].up)
    { Plyr.pos = Map.Stairs[i].pos;
      break;
    }
  }
  Con.gotoXY(Plyr.pos.x, Plyr.pos.y+TOPAREA);
  return 0;
}

Item * Level::addItem(int x, int y, Item *item)
{ Item *it = ::addItem(Items+y*MAPWID + x, item);
  if(it) it->pos = Coord(x, y);
  return it;
}

Item * Level::remItem(int x, int y, Item *item, int count)
{ return ::remItem(Items+y*MAPWID + x, item, count);
}

Coord Level::freeSpace(bool roomonly)
{ return Map.freeSpace(roomonly);
}

void Level::render()
{ Map.render();
}

void Level::draw()
{ Map.draw();
  int i;

  for(i=0;i<MAPWID*MAPHEI;i++)
    if(Items[i] && Map.DisVis[i])
      Con.bothTo(Items[i]->pos.x, Items[i]->pos.y+TOPAREA, ItemChars[Items[i]->type]);

  for(i=0;i<numCreats;i++)
  { if(Map.DisVis[Creats[i].pos.y*MAPWID+Creats[i].pos.x])
      Con.bothTo(Creats[i].pos.x, Creats[i].pos.y+TOPAREA, CreatChars[Creats[i].type]);
  }
  Con.bothTo(Plyr.pos.x, Plyr.pos.y+TOPAREA, CreatChars[Plyr.type]);
  Con.gotoXY(Plyr.pos.x, Plyr.pos.y+TOPAREA);
}
