#ifndef _MAP_H
#define _MAP_H

#include <stdio.h>
#include <assert.h>
#include "enums.h"
#include "datatype.h"
#include "config.h"

struct CMap
{ CMap();
  ~CMap();

  int   init();
  void  deinit();
  int   load(FILE *);                 // returns 0 on success
  int   save(FILE *);

  void  render();
  void  draw();

  Coord freeSpace(bool roomonly);

  byte  layAt(int x, int y)           // returns layer byte at x,y
  { if(x<0 || y<0 || x>=MAPWID || y>=MAPHEI) return MT_NONE;
    return Layout[y*MAPWID+x]; }
  byte  layTo(int x, int y, byte b)   // returns byte passed
  { return Layout[y*MAPWID+x] = b; }
  int   layAdj(int x, int y, byte b); // returns number of b's adjacent to x,y (8-way adjacent)

  byte passable(int x, int y)         // returns true if you can step here
  { return isPassable(layAt(x, y)); }
  byte door(int x, int y)             // returns true if x,y is a door
  { return isDoor(layAt(x, y)); }

  int doorAdj(int x, int y);          // returns number of doors adjacent to x,y

  enum
  { ISPASSABLE=1, ISWALL=2, ISDOOR=4
  };

  static byte isDoor(byte b)          // returns type of door if b is a door, else NONE
  { if(!(MapFlags[b]&ISDOOR)) return NONE;
    return MapFlags[b]&ISPASSABLE ? D_OPEN : D_CLOSED;
  }
  static byte isStair(byte b)         // returns type of stair if b is a stair, else NONE
  { if(b==MT_STAIRDOWN) return S_DOWN;
    if(b==MT_STAIRUP) return S_UP;
    return NONE;
  }
  static byte isPassable(byte b)      // returns true if you can walk on/in layer type b
  { return MapFlags[b] & ISPASSABLE;
  }
  static byte isWall(byte b)          // returns type of wall if b is a wall, else NONE
  { if(!(MapFlags[b]&ISWALL)) return NONE;
    if(b == MT_HORZ) return W_HORZ;
    if(b == MT_VERT) return W_VERT;
    if(b == MT_UPLEFT  || b == MT_UPRIGHT || b == MT_DNLEFT ||
       b == MT_DNRIGHT)
          return W_CORNER;
    if(b == MT_CROSS || b == MT_UPT || b == MT_DOWNT || b == MT_LEFTT ||
       b == MT_RIGHTT)
          return W_CROSS;
    return NONE;
  }

  int    addRoom();
  int    addStairs(bool up);
  void   updateVis(Creature *);
  void   traceLine(float sx, float sy, float dx, float dy, float dx2, float dy2);

  byte   Layout[MAPWID*MAPHEI], DisVis[MAPWID*MAPHEI], Lit[MAPWID*MAPHEI];
  byte   Outbuf[MAPWID*MAPHEI*2];

  Room  *Rooms;
  int    numRooms, maxRooms;

  Stair *Stairs;
  int    numStairs, maxStairs;

  static CharData MapChars[], LitMapChars[];
  static byte     MapFlags[];
};

#endif
