#include "chrono.h"

static bool canDig(int x, int y)
{ byte b = Lev.Map.layAt(x, y);
  if(!b || Lev.Map.isPassable(b)) return true;
  b = Lev.Map.isWall(b);
  if(b && b == W_CORNER || b == W_CROSS || Lev.Map.doorAdj(x, y)) return false;
  return true;
}

static void dig(int x, int y)
{ byte b = Lev.Map.layAt(x, y);
  if(b)
  { if(Lev.Map.isPassable(b) || Lev.Map.isDoor(b)) return;
    if(b=Lev.Map.isWall(b)) Lev.Map.layTo(x, y, b==W_VERT?MT_VCLOSED:MT_HCLOSED);
    return;
  }
  Lev.Map.layTo(x, y, MT_COORIDOR);
}

static int tryDig(Coord *c, Coord c2, int dir)
{ int x = c->x+DirMap[dir].x, y = c->y+DirMap[dir].y;
  
  if(!canDig(x, y))
  { x = c->x, y = c->y;
    if(dir&2) // left/right
    { if(c->y < c2.y)
      { if(canDig(x, y+1)) y++;
        else
          return 1;
      }
      else if(canDig(x, y-1)) y--;
      else
        return 1;
    }
    else // up/down
    { if(c->x < c2.x)
      { if(canDig(x+1, y)) x++;
        else
          return 1;
      }
      else if(canDig(x-1, y)) x--;
      else
        return 1;
    }
  }
  dig(c->x=x, c->y=y);
  return 0;
}

static int conRooms(int rm1, int rm2)
{ Coord c1 = Lev.Map.Rooms[rm1].center(), c2 = Lev.Map.Rooms[rm2].center(), save;
  int fail, rnd=0;
  while(rnd || c1.x != c2.x && c1.y != c2.y)
  { if(rnd)
    { c2  = save;
      rnd = 0;
    }
    else if(!Rnd.Rand(4) && Lev.Map.layAt(c1.x, c1.y) != MT_ROOMFLOOR)
    { save = c2;
      c2   = Coord(c1.x+Rnd.Rand(-1, 1), c1.y+Rnd.Rand(-1, 1));
      rnd  = true;
    }
    int xd = c2.x-c1.x, yd = c2.y-c1.y, dir;
    if(abs(xd) > abs(yd)) dir = xd < 0 ? LEFT : RIGHT;
    else dir = yd < 0 ? UP : DOWN;
    if(tryDig(&c1, c2, dir))
    { if(++fail == 2) return 1;
      if(rnd) continue;
      Coord temp=c1;
      c1=c2, c2=temp;
    }
    else fail=0;
  }
  Lev.Map.Rooms[rm1].con = Lev.Map.Rooms[rm2].con = true;
  return 0;
}

/*static Coord addRoomLay(int r, byte b, bool con=false)
{ Room  rm  = Lev.Map.Rooms[r];
  if(!rm.con) return Coord(-1, -1);

  for(int i=0;i<MGTRIES;i++)
  { int x = Rnd.Rand(rm.area.left+1, rm.area.right-1),
        y = Rnd.Rand(rm.area.top+1, rm.area.bottom-1);
    if(Lev.Map.layAt(x, y) == MT_ROOMFLOOR)
    { Lev.Map.layTo(x, y, b);
      return Coord(x, y);
    }
  }
  return Coord(-1, -1);
}*/

int GenDungeon(int minrooms, int maxrooms)
{ if(!Lev.Map.Rooms) Lev.Map.Rooms = new Room[Lev.Map.maxRooms=10];
  Lev.Map.numRooms = 0;

  while(Lev.Map.numRooms<minrooms) if(Lev.Map.addRoom()) return 1;
  while(Lev.Map.numRooms<maxrooms) if(Lev.Map.addRoom()) break;
  
  int r;
  for(r=0;r<Lev.Map.numRooms;r++)
  { bool  lit = Rnd.Rand(DARKROOM-1) != 0;
    Room  rm  = Lev.Map.Rooms[r];
    byte *lay = Lev.Map.Layout+rm.area.top*MAPWID+rm.area.left;
    byte *lgt = Lev.Map.Lit+rm.area.top*MAPWID+rm.area.left;
    int   wid = rm.width()-1, hei = rm.height()-1, y;
    
    for(y=-1;y<hei+1;y++)
    { memset(lgt, lit, wid+1);
      lgt += MAPWID;
    }

    lay[0] = MT_UPLEFT, lay[wid] = MT_UPRIGHT;
    memset(lay+1, MT_HORZ, wid-1);
    lay += MAPWID;

    for(y=1; y<hei; y++)
    { lay[0] = lay[wid] = MT_VERT;
      memset(lay+1, MT_ROOMFLOOR, wid-1);
      lay += MAPWID;
    }

    lay[0] = MT_DNLEFT, lay[wid] = MT_DNRIGHT;
    memset(lay+1, MT_HORZ, wid-1);
  }

  for(int i=1;i<Lev.Map.numRooms;i++) conRooms(i-1, i);
  if(Lev.Map.addStairs(true)) return 2;
  if(Lev.Map.addStairs(false)) return 3;

  return 0;
}
