#include "chrono.h"
#include <stdio.h>
#include <stdarg.h>
#include <memory.h>
#include <string.h>

bool bTxtOnTop;

#include <pshpack4.h>
struct Entry
{ char *str;
  int   len;
};
#include <poppack.h>
static Entry Buffer[BUFLINES];

static void addBuffer(const char *str)
{ Entry temp = Buffer[0];
  memmove(Buffer, Buffer+1, sizeof(Buffer)-sizeof(Entry));
  Buffer[BUFLINES-1] = temp;
  int len = strlen(str);
  if(Buffer[BUFLINES-1].len < len)
  { delete[] Buffer[BUFLINES-1].str;
    Buffer[BUFLINES-1].str = new char[Buffer[BUFLINES-1].len = (len+64) & ~63];
  }
  memcpy(Buffer[BUFLINES-1].str, str, len+1);
}

void saveCursor(bool save)
{ static int x, y;
  if(save) Con.fromXY(&x, &y);
  else Con.gotoXY(x, y);
}

void printLine(const char *str, bool restcurs, RECT *wnd)
{ char  buf[MAXLINE];
  int   i=0, j=0, y, xs, ys, xw, ye;
  char *found;

  if(wnd)
  { xs = wnd->left, ys = wnd->top;
    xw = wnd->right-wnd->left+1, ye = wnd->bottom;
  }
  else xs = ys = 0, xw = Con.Width, ye = TOPAREA-1;
  y = ys;

  if(restcurs) saveCursor();
  clearOut(wnd);
  Con.gotoXY(xs, ys);

  while(true)
  { found = strpbrk(str+j, " \n\t");
    if(!found)
    { Con.putS(str+i);
      break;
    }
    if(found-str-i > xw)
    { if(y==ye)
      { Con.putS("\t-more-");
        byte c = Con.getch();
        if(c==27) break;
        clearOut(wnd);
        Con.gotoXY(xs, y=ys);
      }
      memcpy(buf, str+i, j-i);
      buf[j-i] = 0;
      trim(buf);
      Con.putS(buf);
      Con.gotoXY(xs, ++y);
      i=j;
    }
    else j = found+strspn(found, " \n\t")-str;
  }
  if(restcurs) restCursor();
  bTxtOnTop = true;
}
    
void clearOut(RECT *rect)
{ if(rect) Con.fill(rect->left, rect->top, rect->right-rect->left+1,
                    rect->bottom-rect->top+1, ' ', GREY);
  else
  { Con.fill(0, 0, Con.Width, TOPAREA, ' ', GREY);
    bTxtOnTop = false;
  }
}

const char * recall(bool first)
{ static int index;
  if(first) index = BUFLINES-1;
  if(index<0) return NULL;
  return Buffer[index--].str;
}

void print(const char *str, ...)
{ va_list list;
  va_start(list, str);
  printv(str, list);
  va_end(list);
}

void printv(const char *str, va_list list)
{ char buf[MAXLINE];
  vsprintf(buf, str, list);
  print1(buf);
}

void print1(const char *str)
{ addBuffer(str);
  printLine(str);
}

void printw(RECT *rect, const char *str, ...)
{ char buf[MAXLINE];
  va_list list;
  va_start(list, str);
  vsprintf(buf, str, list);
  va_end(list);
  printLine(buf, true, rect);
}
